from concurrent.futures import Executor, ThreadPoolExecutor

import grpc
from grpc._interceptor import _Channel

from xdcs import object_repository
from xdcs.cmd import Command
from xdcs.config import load_config, MissingConfigurationException
from xdcs.object_repository import ObjectRepository
from xdcs.servicers.Servicers import Servicers
from xdcs.utils.rforward import rforward


class _XDCS:
    _NO_ARG = object()

    _token: str
    _channel: _Channel
    _executor: Executor
    _config: dict
    _or: ObjectRepository

    _fs_repo_path: str
    _or_path: str

    def __init__(self) -> None:
        self._config = load_config()
        self._executor = ThreadPoolExecutor(max_workers=self.config('app.executors', 10))
        self._fs_repo_path = self.config('app.fs_repo_path', './data')
        self._or_path = self._fs_repo_path + '/objects'
        self._or = object_repository.from_path(self._or_path)

    def executor(self) -> Executor:
        return self._executor

    def execute(self, command: Command) -> None:
        command.execute()

    def run(self) -> None:
        local_port = self.config('app.local_port', 25254)
        server_host = self.config('server.host')
        server_port = self.config('server.port.ssh')

        server = grpc.server(self.executor())
        Servicers.register_all(server)
        server.add_insecure_port('0.0.0.0:' + str(local_port))
        server.start()
        with rforward(local_port, server_host, server_port, False):
            pass

    def set_token(self, token: str) -> None:
        self._token = token

    def token(self) -> str:
        return self._token

    def set_channel(self, channel) -> None:
        self._channel = channel

    def object_repository(self) -> ObjectRepository:
        return self._or

    def channel(self):
        return self._channel

    def config(self, prop: str, default_value=_NO_ARG):
        props = prop.split('.')
        current_config = self._config

        for p in props:
            if p not in current_config:
                if default_value == self._NO_ARG:
                    raise MissingConfigurationException('Required config: ' + prop)

                return default_value

            current_config = current_config[p]

        return current_config



import os
import shutil
import tempfile
from os import path
from typing import IO, Generator, List
from zipfile import ZipFile

from xdcs.app import xdcs
from xdcs.cmd import Command
from xdcs_api.common_pb2 import Chunk
from xdcs_api.object_repository_pb2 import ObjectIds, DependencyResolutionRequest, ObjectKey, ObjectType
from xdcs_api.object_repository_pb2_grpc import ObjectRepositoryStub


class FetchDeploymentCmd(Command):
    _deployment_id: str

    def __init__(self, deployment_id: str):
        self._deployment_id = deployment_id

    def execute(self):
        stub = ObjectRepositoryStub(xdcs().channel())
        req = DependencyResolutionRequest()
        deployment_key = ObjectKey()
        deployment_key.objectId = self._deployment_id
        deployment_key.objectType = ObjectType.DEPLOYMENT
        req.objectKeys.extend([deployment_key])
        req.depth = 2 ** 32 - 1
        object_ids = stub.ResolveDependencies(req).objectIds
        object_ids.append(self._deployment_id)

        xdcs().execute(RetrieveObjectsCmd((ids for ids in [object_ids])))


class RetrieveObjectsCmd(Command):
    _object_ids: Generator[List[str], None, None]

    def __init__(self, object_ids: Generator[List[str], None, None]):
        self._object_ids = object_ids

    def execute(self):
        stub = ObjectRepositoryStub(xdcs().channel())
        chunks = stub.RequestObjects(self.__generate_requests())

        tmp_dir = tempfile.mkdtemp()
        try:
            zip_path = tmp_dir + '/zip'
            # download the zip file
            with open(zip_path, 'wb') as zipfile:
                self.__write_chunks(zipfile, chunks)

            # extract objects
            with ZipFile(zip_path) as zipfile:
                zipfile.extractall(tmp_dir)
                objects = zipfile.namelist()

            # remove the zip file
            os.remove(zip_path)

            obj_repo = xdcs().object_repository()
            for obj in objects:
                obj_repo.import_object(tmp_dir + '/' + obj, required_id=obj)
        finally:
            shutil.rmtree(tmp_dir)

    def __write_chunks(self, zipfile: IO, chunks):
        for chunk in chunks:
            chunk: Chunk
            zipfile.write(chunk.content)

    def __generate_requests(self):
        for ids in self._object_ids:
            requested_objects = ObjectIds()
            requested_objects.objectIds.extend(ids)
            yield requested_objects


class DumpObjectRepositoryTreeCmd(Command):
    _root_id: str
    _out_path: str

    def __init__(self, root_id: str, out_path: str):
        self._root_id = root_id
        self._out_path = out_path

    def execute(self):
        self._dump_tree(self._root_id, self._out_path)

    def _dump_tree(self, root_id, out_path):
        entries = xdcs().object_repository().cat_json(root_id)
        for entry in entries:
            file_path = path.join(out_path, entry['name'])
            mode = entry['mode']
            object_id = entry['id']

            object_type = mode[:2]
            permissions = int(mode[2:], 8)

            if object_type == '12':
                content = xdcs().object_repository().cat_bytes(object_id)
                os.symlink(content, file_path)
                os.chmod(file_path, permissions)
            elif object_type == '10':
                content = xdcs().object_repository().cat_bytes(object_id)
                with open(file_path, 'wb+') as fh:
                    fh.write(content)
                os.chmod(file_path, permissions)
            elif object_type == '04':
                os.mkdir(file_path, permissions)
                self._dump_tree(object_id, file_path)



import hashlib
import json
import os
import re
import shutil
from os import path


class ObjectRepositoryException(Exception):
    pass


class ObjectRepository:
    _object_id_pattern = re.compile("^[0-9a-z]{40}$")

    _path: str

    def __init__(self, path: str) -> None:
        self._path = path

    def import_object(self, path: str, required_id: str = None) -> str:
        object_id = self.__hash(path)

        if required_id is not None and required_id != object_id:
            raise ObjectRepositoryException('Import failed: wrong checksum for {}'.format(path))

        object_path = self.__to_path(object_id, True)
        shutil.copyfile(path, object_path)

        return object_id

    def __hash(self, path: str) -> str:
        sha1 = hashlib.sha1()

        with open(path, 'rb') as f:
            while True:
                data = f.read(4 * 1024)
                if not data:
                    break
                sha1.update(data)

        return sha1.hexdigest()

    def __to_path(self, object_id: str, create_dirs=False) -> str:
        if not self._object_id_pattern.match(object_id):
            raise ObjectRepositoryException('Invalid object ID: ' + object_id)

        containing_dir = path.join(self._path, object_id[:2])

        if create_dirs:
            os.makedirs(containing_dir, exist_ok=True)

        return path.join(containing_dir, object_id[2:])

    def cat_bytes(self, object_id: str) -> bytes:
        with open(self._path + '/' + object_id[:2] + '/' + object_id[2:], "rb") as f:
            return f.read()

    def cat_json(self, object_id: str) -> dict:
        return json.loads(self.cat_bytes(object_id))


def from_path(path: str) -> ObjectRepository:
    return ObjectRepository(path)



import logging
import os

import toml as toml

_config_locations = [
    './xdcs-agent.toml',
    './xdcs-agent.conf',
    '/etc/xdcs/xdcs-agent.toml',
    '/etc/xdcs/xdcs-agent.conf',
]

logger = logging.getLogger(__name__)


class ConfigurationNotFoundException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class MissingConfigurationException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def load_config():
    for config_loc in _config_locations:
        logger.debug('Checking location ' + config_loc + ' for configuration')
        if os.path.isfile(config_loc):
            return toml.load(config_loc)
        else:
            logger.debug('Config not found in ' + config_loc)

    raise ConfigurationNotFoundException(
        'Configuration not found, searched locations: ' + str(_config_locations))



import logging

from xdcs.app import xdcs

logger = logging.getLogger(__name__)


def asynchronous(func):
    def task(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            logger.exception('Exception during asynchronous execution: ' + str(e))
            raise e

    def replacement(*args, **kwargs):
        return xdcs().executor() \
            .submit(lambda: task(*args, **kwargs))

    return replacement



import logging
import tempfile
from os import path

from xdcs.app import xdcs
from xdcs.cmd import Command
from xdcs.cmd.object_repository import FetchDeploymentCmd, DumpObjectRepositoryTreeCmd
from xdcs.docker import DockerCli

logger = logging.getLogger(__name__)


class RunTaskCmd(Command):
    _deployment_id: str

    def __init__(self, deployment_id: str) -> None:
        self._deployment_id = deployment_id

    def execute(self):
        logger.info('Running a deployment: ' + self._deployment_id)
        with tempfile.TemporaryDirectory() as workspace_path:
            xdcs().execute(FetchDeploymentCmd(self._deployment_id))
            deployment: dict = xdcs().object_repository().cat_json(self._deployment_id)
            root_id = deployment['root']
            xdcs().execute(DumpObjectRepositoryTreeCmd(root_id, workspace_path))
            config_type = deployment['config']['type']
            if config_type == 'docker':
                xdcs().execute(RunDockerTaskCmd(workspace_path, deployment, self._deployment_id))
            else:
                raise Exception('Unsupported task type ' + config_type)
        logger.info('Deployment finished: ' + self._deployment_id)


class RunDockerTaskCmd(Command):
    _deployment_id: str
    _deployment: dict

    def __init__(self, workspace_path: str, deployment: dict, deployment_id: str) -> None:
        self._workspace_path = workspace_path
        self._deployment_id = deployment_id
        self._deployment = deployment

    def execute(self):
        deployment = self._deployment
        dockerfile = deployment['config'].get('dockerfile', None)

        if dockerfile is None or len(dockerfile) == 0:
            dockerfile = 'Dockerfile'

        dockerfile = path.join(self._workspace_path, dockerfile)
        image_id = DockerCli().build(self._workspace_path, dockerfile)
        logger.info('Docker built, image ID: ' + image_id)

        docker_cli = DockerCli()
        docker_cli.remove_container_after_finish()
        docker_cli.container_name('xdcs_' + self._deployment_id)
        docker_cli.run(image_id)



import logging
import re
import subprocess
from typing import Union, List

from lazy import lazy
from packaging import version

from xdcs import gpu
from xdcs.app import xdcs

logger = logging.getLogger(__name__)


class DockerException(Exception):
    pass


class DockerCli:
    _docker_exec: str
    _opts: [str] = []
    _nvidia_all_devices: bool = False
    _nvidia_device_ids: [str] = []

    def __init__(self) -> None:
        self._docker_exec = xdcs().config('system.docker', 'docker')

    def opt(self, opt: Union[str, List[str]]) -> None:
        if isinstance(opt, str):
            self._opts.extend([opt])
        else:
            self._opts.extend(opt)

    def remove_container_after_finish(self) -> None:
        self._opts.extend(['--rm'])

    def container_name(self, name: str) -> None:
        self._opts.extend(['--name', name])

    def nvidia_all_devices(self):
        self._nvidia_all_devices = True

    def nvidia_device(self, key: str) -> None:
        device = gpu.manager.device_by_key(key)
        if not device.is_nvidia():
            raise DockerException('Non-Nvidia devices are not supported by Docker')

        self._nvidia_device_ids.extend(device.nvidia_id())

    def __run_docker(self, args: [str]) -> subprocess.CompletedProcess:
        return subprocess.run([self._docker_exec, *args],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              text=True)

    def run(self, image: str) -> None:
        opts = [*self._opts, *self.__build_gpu_opts()]
        proc = subprocess.Popen([self._docker_exec, 'run', *opts, image],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        for line in proc.stdout:
            logger.info(line.decode("utf-8").rstrip('\n'))

        proc.wait()

        if proc.returncode != 0:
            raise DockerException('Docker run failed: ' + str(proc.stderr.read().decode("utf-8")))

    def build(self, directory, dockerfile) -> str:
        proc = self.__run_docker(['build', '-q', '-f', dockerfile, directory])
        if proc.returncode != 0:
            raise DockerException('Docker build failed: ' + str(proc.stderr))
        return proc.stdout.strip()

    def version(self) -> str:
        proc = self.__run_docker(['--version'])
        if proc.returncode != 0:
            raise DockerException('Docker version retrieval failed: ' + str(proc.stderr))
        return proc.stdout.strip()

    def __build_gpu_opts(self):
        device_list_str = ','.join(self._nvidia_device_ids)

        if info.is_native_gpu_supported():
            if self._nvidia_all_devices:
                return ['--gpus', 'all']

            if self._nvidia_device_ids:
                return ['--gpus', 'device=' + device_list_str]

            return []

        if info.is_native_gpu_supported():
            if self._nvidia_all_devices:
                return ['--runtime=nvidia',
                        '-e', 'NVIDIA_VISIBLE_DEVICES=all']

            if self._nvidia_device_ids:
                return ['--runtime=nvidia',
                        '-e', 'NVIDIA_VISIBLE_DEVICES=' + device_list_str]

            return []

        return []


class _DockerInfo:
    @lazy
    def docker_available(self) -> bool:
        try:
            DockerCli().version()
            return True
        except DockerException:
            return False

    @lazy
    def version(self) -> str:
        docker_version_output = DockerCli().version()
        version_search = re.search('Docker version ([0-9.]+), build ([a-f0-9]+)',
                                   docker_version_output, re.IGNORECASE)

        if version_search:
            return version_search.group(1)
        else:
            raise DockerException('Unrecognized version output: ' + docker_version_output)

    @lazy
    def is_native_gpu_supported(self) -> bool:
        if version.parse(self.version()) < version.parse("19.03"):
            return False

        try:
            DockerCli() \
                .opt(['--gpus', 'all']) \
                .run('hello-world')
            return True
        except DockerException as e:
            if 'could not select device driver' in str(e):
                return False
            raise e

    @lazy
    def is_nvidia_runtime_available(self) -> bool:
        try:
            DockerCli() \
                .opt('--runtime=nvidia') \
                .run('hello-world')
            return True
        except DockerException as e:
            if 'Unknown runtime' in str(e):
                return False
            raise e


info = _DockerInfo()



import os

_config_locations = [
    'conf/xdcs-agent.toml',
    'conf/xdcs-agent.conf',
    './conf/xdcs-agent.toml',
    './conf/xdcs-agent.conf',
]


def main():
    for config_loc in _config_locations:
        print('Checking location ' + config_loc + ' for configuration')
        if os.path.isfile(config_loc):
            print('Found in ' + config_loc)



from enum import Enum


class KernelParameter:
    def __init__(self, direction, value, position, param_type, name):
        self.direction = direction
        self.value = value
        self.position = position
        self.type = param_type
        self.name = name

    def get_type(self):
        return self.type

    def get_value(self):
        return self.value


class ParameterDirection(Enum):
    IN = 1
    OUT = 2
    IN_OUT = 3


class KernelParameters:
    def __init__(self, *args):
        self.param_list = list(args)

    def get_params_values(self):
        param_list = list()
        for param in self.param_list:
            param_list.append(param.get_value())
        return param_list


class ParameterType(Enum):
    SIMPLE = 1
    POINTER = 2



from xdcs.app import xdcs
from xdcs.cmd import Command
from xdcs_api.registration_pb2 import AgentRegistrationRequest
from xdcs_api.registration_pb2_grpc import AgentRegistrationStub


class RegistrationFailedException(Exception):
    pass


class RegisterAgentCmd(Command):
    def execute(self):
        request = AgentRegistrationRequest()
        request.displayName = xdcs().config('agent.name')

        stub = AgentRegistrationStub(xdcs().channel())
        response = stub.Register(request)

        if not response.success:
            raise RegistrationFailedException()



from typing import Dict, Optional


class GpuDevice:
    def nvidia_id(self) -> str:
        return ''

    def is_nvidia(self) -> bool:
        return False


class _GpuManager:
    _devices_by_key: Dict[str, GpuDevice]

    def __init__(self) -> None:
        pass

    def device_by_key(self, key: str) -> Optional[GpuDevice]:
        return self._devices_by_key[key]


manager = _GpuManager()



#!/usr/bin/env python

from setuptools import setup
from setuptools.command.install import install as _install

class install(_install):
    def pre_install_script(self):
        pass

    def post_install_script(self):
        pass

    def run(self):
        self.pre_install_script()

        _install.run(self)

        self.post_install_script()

if __name__ == '__main__':
    setup(
        name = 'xdcs-agent',
        version = '0.1.0.dev20191125215215',
        description = '',
        long_description = '',
        author = 'Kamil Jarosz, Krystian Życiński, Adam Szczerba, Jan Rodzoń',
        author_email = 'kjarosz@student.agh.edu.pl, zycinski@student.agh.edu.pl, adamszczerba19@gmail.com, rodzonjan@wp.pl',
        license = 'GNU GPLv3 ',
        url = 'https://github.com/xdcs-team/xdcs',
        scripts = [],
        packages = [
            'xdcs',
            'xdcs.cmd'
        ],
        namespace_packages = [],
        py_modules = ['__main__'],
        classifiers = [
            'Development Status :: 3 - Alpha',
            'Programming Language :: Python'
        ],
        entry_points = {
            'console_scripts': ['xdcs-agent = xdcs.agent_cli:main']
        },
        data_files = [
            ('/usr/config', ['xdcs/conf/xdcs-agent.toml'])
        ],
        package_data = {},
        install_requires = [
            'distro',
            'lazy',
            'packaging',
            'py-cpuinfo',
            'pyopencl'
        ],
        dependency_links = [],
        zip_safe = True,
        cmdclass = {'install': install},
        keywords = '',
        python_requires = '',
        obsoletes = [],
        include_package_data=True,
    )
